/*
   Yamaha Song Display

   Get the song details from the R-N602 and display them.
   Assumes a 2x16 LCD display.
   Uses pins A1 and A2  to set GMT/BST and clock choice
   the default is GMT and show song details--the time is shown if the R-N602 is switched off

 */

// Configuration
// serverAddress is the IP address of the R-N602
const char serverAddress[] =  "192.168.0.22";
// trackDelay is the length of time the track is displayed in ms
const unsigned int trackDelay = 10000;
// albumDelay is the length of time the album and artist are displayed in ms
const unsigned int artistDelay = 5000;
// scrollDelay is the length of time between shifts when scrolling in ms
const unsigned int scrollDelay = 400;
// End of Configuration 
// Do not change below this line

#define YAMAHA_SONG_SPLASH  "Yamaha Song v1.5"
// JSON
#include <ArduinoHttpClient.h>
#include <ArduinoJson.h>

// Time
#include <NTPClient.h>
#include <TimeLib.h>
#include <functional>
#include <array>
const int TIME_DELAY = 20000; // time the NTP server is polled in ms
// jumper DST_PIN LOW for BST
const int DST_PIN = A1; // A1 is exposed on the LCD Keypad Shield
// jumper TIME_ONLY_PIN LOW to only show the time (no song details)
const int TIME_ONLY_PIN = A2; // A2 is exposed on the LCD Keypad Shield
auto isDaylightSavingTime = []() -> bool {
   return digitalRead(DST_PIN) == LOW; // DST if the pin is LOW
};

auto showTimeOnly = []() -> bool {
   return digitalRead(TIME_ONLY_PIN) == LOW; // Show time only if the pin is LOW
};

// LCD
#include <LiquidCrystal.h>;

LiquidCrystal lcd(8, 9, 4, 5, 6, 7); // select the pins used on the LCD panel
const int LCDwidth = 16;
const int maxItemLength = 64;

// Wi-Fi
#include <WiFiS3.h>
#include <WiFiUDP.h>

int status = WL_IDLE_STATUS;

int httpPort = 80;

WiFiClient wifiClient;
HttpClient songClient = HttpClient(wifiClient, serverAddress, httpPort);

// Time
WiFiUDP ntpUDP;
NTPClient timeClient(ntpUDP, "uk.pool.ntp.org");

void printWifiStatus() {
   // print the SSID of the network you're attached to:
   Serial.print("SSID: ");
   Serial.println(WiFi.SSID());

   // print your board's IP address:
   IPAddress ip = WiFi.localIP();
   Serial.print("IP Address: ");
   Serial.println(ip);

   // print the received signal strength:
   long rssi = WiFi.RSSI();
   Serial.print("signal strength (RSSI):");
   Serial.print(rssi);
   Serial.println(" dBm");
}

void scrollText(int line, const char * text) {
   char lineBuf[LCDwidth + 1];
   size_t len = strlen(text);

   if (len > LCDwidth) {
      for (size_t i = 0; i < len - LCDwidth + 1; i++) { 
         lcd.setCursor(0, line); 
         memset(lineBuf, 0, sizeof(lineBuf));
         strncpy(lineBuf, text + i, sizeof(lineBuf) - 1);
         lcd.print(lineBuf); 
         delay(scrollDelay); 
      } 
   } 

   lcd.setCursor(0, line); 
   memset(lineBuf, 0, sizeof(lineBuf));
   strncpy(lineBuf, text, sizeof(lineBuf) - 1);
   lcd.print(lineBuf);
   delay(scrollDelay); 
}

auto fetchPlayInfo = []() -> String {
   if (WiFi.status() != WL_CONNECTED) {
      Serial.println("WiFi not connected");
      return "";
   }
   songClient.get("/YamahaExtendedControl/v1/netusb/getPlayInfo");
   int statusCode = songClient.responseStatusCode();
   if (statusCode != 200) {
      Serial.print("Song Status code: ");
      Serial.println(statusCode);
      return "";
   }
   return songClient.responseBody();
};

auto parseJSON = [](const String& payload) -> StaticJsonDocument<1024> {
   StaticJsonDocument<1024> doc;
   DeserializationError error = deserializeJson(doc, payload);
   if (error) {
      Serial.print("deserializeJson() failed: ");
      Serial.println(error.f_str());
      return {};
   }
   return doc;
};

auto displayText = [](const char* text1, const char* text2) {
   lcd.clear();
   lcd.setCursor(0, 0);
   lcd.print(text1);
   scrollText(1, text2);
};

auto alternate = []() -> bool {
   static int current = 1; // Start with an odd number
   current += 1; // Increment to the next number
   return current % 2 == 0;
};	

auto UTF8toASCIImap = []() -> std::vector<std::pair<const char *, char>> {
   return {
      {"\xC3\xA0", 'a'}, {"\xC3\xA1", 'a'}, {"\xC3\xA2", 'a'}, {"\xC3\xA3", 'a'},
         {"\xC3\xA4", 'a'}, {"\xC3\xA5", 'a'}, {"\xC3\xA8", 'e'}, {"\xC3\xA9", 'e'},
         {"\xC3\xAA", 'e'}, {"\xC3\xAB", 'e'}, {"\xC3\xAC", 'i'}, {"\xC3\xAD", 'i'},
         {"\xC3\xAE", 'i'}, {"\xC3\xAF", 'i'}, {"\xC3\xB2", 'o'}, {"\xC3\xB3", 'o'},
         {"\xC3\xB4", 'o'}, {"\xC3\xB5", 'o'}, {"\xC3\xB6", 'o'}, {"\xC3\xB9", 'u'},
         {"\xC3\xBA", 'u'}, {"\xC3\xBB", 'u'}, {"\xC3\xBC", 'u'}, {"\xC3\xB1", 'n'},
         {"\xC3\xA7", 'c'}, {"\xC3\x80", 'A'}, {"\xC3\x81", 'A'},
         {"\xC3\x82", 'A'}, {"\xC3\x83", 'A'}, {"\xC3\x84", 'A'}, {"\xC3\x85", 'A'},
         {"\xC3\x88", 'E'}, {"\xC3\x89", 'E'}, {"\xC3\x8A", 'E'}, {"\xC3\x8B", 'E'},
         {"\xC3\x8C", 'I'}, {"\xC3\x8D", 'I'}, {"\xC3\x8E", 'I'}, {"\xC3\x8F", 'I'},
         {"\xC3\x92", 'O'}, {"\xC3\x93", 'O'}, {"\xC3\x94", 'O'}, {"\xC3\x95", 'O'},
         {"\xC3\x96", 'O'}, {"\xC3\x99", 'U'}, {"\xC3\x9A", 'U'}, {"\xC3\x9B", 'U'},
         {"\xC3\x9C", 'U'}, {"\xC3\x91", 'N'}, {"\xC3\x87", 'C'}
   };
};

// Lambda to Map UTF-8 to ASCII
auto UTF8toASCII = [](const char *utf8Char) -> char {
   // UTF-8 to ASCII Mapping as a Lambda
   auto map = UTF8toASCIImap();
   for (const auto &entry : map) {
      if (strncmp(utf8Char, entry.first, 2) == 0) {
         return entry.second;
      }
   }
   return '?';
};

// Lambda to Convert UTF-8 String to ASCII
auto convertUTF8toASCII = [](char *dest, const char *src, size_t n) {
   size_t i = 0, j = 0;

   while (i < n && src[i] != '\0') {
      if ((src[i] & 0x80) == 0) {
         // ASCII character
         dest[j++] = src[i++];
      } else {
         // UTF-8 character
         char utf8Char[3] = {src[i], src[i + 1], '\0'};
         dest[j++] = UTF8toASCII(utf8Char);
         i += 2;
      }
   }
   dest[j] = '\0'; // Null-terminate
};

auto processDetails = [](const StaticJsonDocument<1024>& doc) -> void {
   const char* artist = doc["artist"];
   const char* track = doc["track"];
   const char* album = doc["album"];


   // body
   char line1[maxItemLength + 1] = {0};
   char line2[maxItemLength + 1] = {0};
   char trackBuf[maxItemLength + 1] = {0};

   if (alternate()) {
      convertUTF8toASCII(line1, artist, maxItemLength);
      convertUTF8toASCII(line2, album, maxItemLength);
      displayText(line1, line2);
      delay(artistDelay);
   } else {
      strncpy(trackBuf, track, maxItemLength);
      convertUTF8toASCII(line1, trackBuf, LCDwidth);
      convertUTF8toASCII(line2, trackBuf + LCDwidth, maxItemLength - LCDwidth);
      displayText(line1, line2);
      delay(trackDelay);
   }
};

auto getSongDetails = [&]() -> int {
   String payload = fetchPlayInfo();
   if (payload.length() == 0) {
      return 1; // Error
   }
   StaticJsonDocument<1024> doc = parseJSON(payload);
   if (doc.isNull()) {
      return 2; // Parsing error
   }
   processDetails(doc);
   return 0;
};

String formatDate(unsigned long epochTime) {
   // Array of day names
   const std::array<const char*, 7> dayNames = {"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"};
   // Array of month names
   const std::array<const char*, 12> monthNames = {"Jan", "Feb", "Mar", "Apr", "May", "Jun", 
      "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};

   // Helper lambdas
   auto daysInMonth = [](int year) -> std::array<int, 12> {
      return {31, (year % 4 == 0 && year % 100 != 0) || (year % 400 == 0) ? 29 : 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
   };

   auto calculateDayOfWeek = [](unsigned long daysSinceEpoch) -> int {
      return (daysSinceEpoch + 4) % 7; // 1970-01-01 was a Thursday
   };

   auto calculateYear = [](unsigned long& daysSinceEpoch) -> int {
      int year = 1970;
      while (daysSinceEpoch >= 365) {
         if ((year % 4 == 0 && year % 100 != 0) || (year % 400 == 0)) {
            if (daysSinceEpoch >= 366) {
               daysSinceEpoch -= 366;
               year++;
            }
         } else {
            daysSinceEpoch -= 365;
            year++;
         }
      }
      return year;
   };

   auto calculateMonthAndDay = [&daysInMonth](unsigned long& daysSinceEpoch, int year) -> std::pair<int, int> {
      int month = 0;
      auto daysInMonthArray = daysInMonth(year);
      while (daysSinceEpoch >= daysInMonthArray[month]) {
         daysSinceEpoch -= daysInMonthArray[month];
         month++;
      }
      int day = daysSinceEpoch + 1;
      return {month, day};
   };

   // Calculate the date components
   unsigned long secondsInDay = 86400;
   unsigned long daysSinceEpoch = epochTime / secondsInDay;

   int dayOfWeek = calculateDayOfWeek(daysSinceEpoch);
   int year = calculateYear(daysSinceEpoch);
   auto [month, day] = calculateMonthAndDay(daysSinceEpoch, year);

   // Format the date string
   char dateStr[20];
   snprintf(dateStr, sizeof(dateStr), "%s %02d %s %04d", dayNames[dayOfWeek], day, monthNames[month], year);

   return String(dateStr);
}

auto getTimeDetails = []() {
   unsigned long epoch;

   if (WiFi.status() == WL_CONNECTED) {
      timeClient.setTimeOffset(isDaylightSavingTime() ? 3600 : 0); // GMT or BST -- other time zones not catered for
      timeClient.update();
      lcd.clear();
      lcd.setCursor(0,0); 
      lcd.print(timeClient.getFormattedTime().substring(0, 5));

      lcd.setCursor(0,1); 
      lcd.print(formatDate(timeClient.getEpochTime()));
   } else {
      Serial.println("Wifi not connected in getTimeDetails");
   }
};

void connectWifi() {
#include "arduino_secrets.h" 
   char ssid[] = SECRET_SSID;    // the network SSID (name)
   char pass[] = SECRET_PASS;    // the network password

   // attempt to connect to WiFi network:
   while (WiFi.status() != WL_CONNECTED) {
      lcd.setCursor(0,1); 
      lcd.print("Connecting ...");
      Serial.print("Attempting to connect to SSID: ");
      Serial.println(ssid);
      WiFi.begin(ssid, pass);

      // wait 10 seconds for connection:
      delay(10000);
   }
   lcd.setCursor(0,1); 
   lcd.print("                ");
   lcd.print(ssid);

   printWifiStatus();
}

void setup() {
   // Open serial communications and wait for port to open:
   Serial.begin(9600);
   while (!Serial) {
      ; // wait for serial port to connect. 
   }
   Serial.println(YAMAHA_SONG_SPLASH);

   // LCD
   lcd.begin(LCDwidth, 2); 
   lcd.setCursor(0,0); 
   lcd.print(YAMAHA_SONG_SPLASH);

   connectWifi(); // hangs on failure as nothing can be done

   timeClient.begin();

   pinMode(DST_PIN, INPUT_PULLUP);
   pinMode(TIME_ONLY_PIN, INPUT_PULLUP);
}

void loop() {
   if (showTimeOnly() || (getSongDetails() != 0)) {
      getTimeDetails();
      delay(TIME_DELAY); // throttle
   }
}
